# (c) 2013-2014 The MITRE Corporation. ALL RIGHTS RESERVED.
# Contributions by Sam Cornwell, John Butterworth, Corey Kallenberg, and Xeno Kovah

"""
Usage:
  bios_diff.py [-crs] [-i IGNORE] [-d [-a [-p]] [-n [-u UNIQUE]] [-l SIZELIMIT] [-m NUMBYTES]] [-o OUT] [-e EFIPWN] <file1> [<file2>]
  bios_diff.py (-h | --help)

  The files are BIOS dumps to be compared.  <file2> may be a single file, or it may be a directory which
  contains several BIOS dumps against which we will diff <file1>.  Also, <file1> can be a directory by itself.  
  In this case, the first file found in this directory will be compared against all of the others.

  
Options:
  -c            delete the directories when the diff is complete
  -r            reuse parsed directories previously generated by EFIPWN if they appear to exist
  -s            print out all sha1 hashes of files in BIOS dump
  -i IGNORE     file containing list of regular expressions (one per line) for filenames we should ignore
  -d            do hash diffing of extracted files
  -a            print all unique ranges per file
  -n            print number of unique bytes per file
  -u UNIQUE     exclude diffs which have less than UNIQUE unique bytes for both files
  -l SIZELIMIT  dont compute unique ranges or bytes on files which exceed this size
  -p            print the PE information about diffs if the files are PE files
  -m NUMBYTES   merge regions which are within NUMBYTES of eachother
  -o OUT        output directory [default: temp]
  -e EFIPWN     the location of EFIPWN files [default: EFIPWN]
  
"""

from docopt import docopt
import os
import shutil
import subprocess
import sys
import filecmp
import mmap
import itertools
import pefile
import glob
import re
import hashlib
import difflib
import re
import multiprocessing
import logging

def sanitize_region((base, limit)):
    if (base > limit):
        return None
    else:
        return base, limit

def region_intersect((base1, limit1), (base2, limit2)):
    sanitize_region((max(base1, base2), min(limit1, limit2)))

def region_cmp((base1, limit1), (base2, limit2)):
    return cmp(base1, base2) if base1 != base2 else cmp(limit1, limit2)

def region_set_union(regions):
    regions = sorted(regions, cmp=region_cmp)
    merged_regions = []
    while len(regions) >= 2:
        base1, limit1 = regions.pop(0)
        base2, limit2 = regions[0]
        
        if (base1 == base2):
            continue
        
        if (limit1 >= base2 - 1):
            regions[0] = (base1, max(limit1, limit2))
        else:
            merged_regions.append((base1, limit1))
        
    if (regions):
        merged_regions.append(regions.pop(0))
    
    return merged_regions

# region1 - region2
# warning: reading may cause aneurysm
def region_diff((base1, limit1), (base2, limit2)):
    if base1 < base2:
        if limit2 < limit1:
            return [(base1, base2 - 1), (limit2 + 1, limit1)]
        else:
            return [(base1, min(base2 - 1, limit1))]
    else:
        if limit2 >= limit1:
            return []
        else:
            return [(max(limit2 + 1, base1), limit1)]

def region_set_diff(region_set, region2):
    return region_set_union(sum([region_diff(region1, region2) for region1 in region_set], []))

def efipwn_dump(bios_dump, out_dir):
    p = subprocess.Popen([sys.executable, os.path.join(args['-e'], 'dump.py'), bios_dump, 'dump', out_dir], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if p.returncode:
        sys.stderr.write(' '.join(["EFIPWN failed for", bios_dump, '\n']))
        sys.stderr.write(stderr)
        return False
    return True

def efipwn_dump_direct(bios_dump, out_dir):
    efipwn_args = [os.path.join(args['-e'], 'dump.py'), bios_dump, 'dump', out_dir]
    try:
        EfipwnDumpScript.main(efipwn_args)
    except Exception, e:
        sys.stderr.write(' '.join(["EFIPWN failed for", bios_dump, '\n']))
        sys.stderr.write(str(e))
        return False
    return True

'''
def efipwn_dump_direct(bios_dump, out_dir):
    try:
        with open(bios_dump, 'rb') as f:
            f.seek(0, os.SEEK_END)
            flen = f.tell()
            f.seek(0, os.SEEK_SET)
            EfiTreeFileDumpVisitor(out_dir).visit(EfiFirmwareImage(f, flen))
            return True
    except Exception, e:
        sys.stderr.write(' '.join(["EFIPWN failed for", bios_dump, '\n']))
        sys.stderr.write(e)
        sys.stderr.write('\n')
        return False
'''

def strip_end_number(pattern, path):
    return re.sub(pattern + '[0-9]+', pattern, path)

def canonicalize_path(path):
    path = strip_end_number('PE32_', path)
    path = strip_end_number('PEI_DEPEX_', path)
    path = strip_end_number('DXE_DEPEX_', path)
    path = strip_end_number('RAW_', path)
    path = strip_end_number('fv', path)
    return path

def get_containing_dicts(key, value, dicts):
    return [d for d in dicts if key in d and d[key] == value]

# NOTE: one subtlety here is that duplicate file hashes in one list get disregarded as 
# long as there is at least one matching hash in the other list
def diff_file_hashes(file_infos1, file_infos2):
    hash_set1 = set(file_info['hash'] for file_info in file_infos1)
    hash_set2 = set(file_info['hash'] for file_info in file_infos2)

    hash_set_diff = hash_set1 - hash_set2
    file_infos = file_infos1
    left = [file_info for file_info in file_infos if file_info['hash'] in hash_set_diff]

    hash_set_diff = hash_set2 - hash_set1
    file_infos = file_infos2
    right = [file_info for file_info in file_infos if file_info['hash'] in hash_set_diff]

    return (left, right)

def get_best_match(one, others):
    matches = get_containing_dicts('canon_path', one['canon_path'], others)
    if not matches and one['uistring']:
        matches = get_containing_dicts('uistring', one['uistring'], others)
    
    if matches:
        return max(matches, key=lambda other: difflib.SequenceMatcher(a=other['path'], b=one['path']).ratio())
    
    return None

def get_uistring(file_info):
    if 'uistring' in file_info.iterkeys():
        return file_info['uistring']

    directory = os.path.split(file_info['path'])[0]
    uistring_path = os.path.join(directory, 'uistring.txt')
    try:
        with open(uistring_path, 'r') as f:
            uistring = f.readline()
    except:
        uistring = None

    file_info['uistring'] = uistring
    return uistring

def print_with_uistring(file_info):
    uistring = get_uistring(file_info)
    if uistring:
        print ''.join([file_info['path'], ' (', uistring, ')'])
    else:
        print file_info['path']

def strip_hex_l(s):
    return s[:-1] if s[-1] == 'L' else s

def get_unique_regions_old(file_info1, file_info2):
    with open(file_info1['path'], 'rb') as file1, open(file_info2['path'], 'rb') as file2:
        data1 = file1.read()
        data2 = file2.read()

        seq = difflib.SequenceMatcher(a=data1, b=data2)

        diffs = []

        for tag, i1, i2, j1, j2 in seq.get_opcodes():
            if tag != 'equal':
                if i1 != i2:
                    diffs.append((file_info1, i1, i2))
                if j1 != j2:
                    diffs.append((file_info2, j1, j2))

        diffs.sort(key=lambda (file_info, start, end): (start, end))

        return diffs

def get_unique_regions(file_info1, file_info2):
    with open(file_info1['path'], 'rb') as file1, open(file_info2['path'], 'rb') as file2:
        data1 = file1.read()
        data2 = file2.read()

        seq = difflib.SequenceMatcher(a=data1, b=data2)

        data_regions1 = [(0, len(data1) - 1)]
        data_regions2 = [(0, len(data2) - 1)]

        matching_blocks = seq.get_matching_blocks()

        for i, j, n in matching_blocks:
            if data_regions1:
                data_regions1.extend(region_diff(data_regions1.pop(), (i, i+n-1)))
            if data_regions2:
                data_regions2.extend(region_diff(data_regions2.pop(), (j, j+n-1)))

        '''
        for data_regions in [data_regions1, data2_regions]:
            for i in xrange(len(data_regions)):
                base, limit = data_regions[i]
                data_regions[i] = (base, limit+1)
        '''

        return data_regions1, data_regions2

def get_separated_unique_regions(file_info1, file_info2):
    with open(file_info1['path'], 'rb') as file1, open(file_info2['path'], 'rb') as file2:
        data1 = file1.read()
        data2 = file2.read()

        seq = difflib.SequenceMatcher(a=data1, b=data2)

        left_diffs = []
        right_diffs = []

        for tag, i1, i2, j1, j2 in seq.get_opcodes():
            if tag != 'equal':
                if i1 != i2:
                    left_diffs.append((i1, i2))
                if j1 != j2:
                    right_diffs.append((j1, j2))

        return left_diffs, right_diffs

def merge_regions(regions, merge_gap_size):
    if not regions:
        return []
    new_regions = [regions[0]]
    for region in itertools.islice(regions, 1, None):
        # base - limit - 1
        if region[0] - new_regions[-1][1] - 1 <= merge_gap_size:
            new_regions[-1] = (new_regions[-1][0], region[1])
        else:
            new_regions.append(region)
    return new_regions

def generate_directories(fle):
    if os.path.isdir(fle):
        directory = fle
    elif os.path.isfile(fle):
        directory = os.path.join(args['-o'], os.path.split(fle)[1])
        if not args['-r'] or not glob.glob(os.path.join(directory, '*')):
            try:
                shutil.rmtree(directory)
            except:
                pass
            
            if not efipwn_dump_direct(fle, directory):
                return None
    else:
        sys.stderr.write(' '.join(['File', fle, 'does not exist.', '\n']))
        return None

    return directory

def generate_file_infos(directory):
    file_infos = []
    for root, subdirs, files in os.walk(directory):
        for fname in files:
            if fname != 'uistring.txt':
                path = os.path.join(root, fname)
                rel_path = os.path.relpath(path, directory)
                canon_path = canonicalize_path(rel_path)

                if any(regex.search(canon_path) for regex in ignore):
                    continue

                with open(path, 'rb') as f:
                    file_info = {}
                    data = f.read()
                    file_info['hash'] = hashlib.sha1(data).hexdigest()
                    file_info['canon_path'] = canon_path
                    file_info['path'] = path
                    get_uistring(file_info)

                    file_infos.append(file_info)
    return file_infos

#http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order?answertab=votes
def uniquify(seq):
    seen = set()
    seen_add = seen.add
    return [ x for x in seq if x not in seen and not seen_add(x)]


def print_pe(file_info, regions):
    for start, end in regions:
        if os.path.split(file_info['path'])[1].startswith("PE32"):
            try:
                pe = pefile.PE(file_info['path'])
                section = pe.get_section_by_offset(start)
                print "PE Information:"
                if not section:
                    print "Diff not in any section"
                else:
                    print "Section " + section.Name
                    rva = section.get_rva_from_offset(start)
                    va = pe.OPTIONAL_HEADER.ImageBase + rva
                    print "RVA " + strip_hex_l(hex(rva))
                    print "VA " + strip_hex_l(hex(va))
            except pefile.PEFormatError:
                pass

def calculate_unique_bytes(regions):
    return reduce(lambda acc, (start, end): acc + (end-start) + 1, regions, 0)

# this magic has to be done outside of the main for the multiprocessing to work
args = docopt(__doc__)

sys.path.append(args['-e'])
import dump as EfipwnDumpScript

if args['-i']:
    with open(args['-i'], 'r') as f:
        ignore = [re.compile(regex.strip()) for regex in f.readlines()]
else:
    ignore = []

if __name__ == '__main__':
    logging.basicConfig(level=logging.CRITICAL)

    if not args['<file2>']:
        if os.path.isdir(args['<file1>']):
            files = glob.glob(os.path.join(args['<file1>'], '*'))
        else:
            files = [args['<file1>']]
    else:
        if os.path.isdir(args['<file2>']):
            files = [args['<file1>']] + glob.glob(os.path.join(args['<file2>'], '*'))
        else:
            files = [args['<file1>'], args['<file2>']]

    files = uniquify(map(os.path.normpath, files))
    
    num_processes = min(len(files), multiprocessing.cpu_count())
    if num_processes > 1:
        pool = multiprocessing.Pool(processes=num_processes)
        do_map = pool.map
    else:
        do_map = map

    #directories = [generate_directories(f, args['-o'], args['-r'], args['-e']) for f in files]
    directories = do_map(generate_directories, files)

    #exclude any files which failed to be extracted
    directories = [d for d in directories if d]

    if not directories:
        sys.exit(1)

    #file_infos = [generate_file_infos(d, ignore) for d in directories]
    file_infos = do_map(generate_file_infos, directories)

    if args['-s']:
        for file_info in itertools.chain.from_iterable(file_infos):
            print file_info['path']
            print file_info['hash']

    if args['-d']:
        directory_left, file_infos_left =  directories[0], file_infos[0]

        for directory_right, file_infos_right in itertools.izip(directories[1:], file_infos[1:]):
            #print ' '.join(['Diffing', directory_left, 'and', directory_right])
            #print # newline

            left_diffs, right_diffs = diff_file_hashes(file_infos_left, file_infos_right)

            diffs = itertools.chain(
                ((one, file_infos_right) for one in left_diffs),
                ((one, file_infos_left) for one in right_diffs)
            )

            # used to prevent printing the same file's info twice
            printed = []

            for one, others in diffs:
                if one not in printed: 
                    other = get_best_match(one, others)

                    # there's a match?
                    if other:
                        printed.append(other)

                        do_regions = args['-n'] or args['-a']
                        if do_regions and args['-l']:
                            do_regions = do_regions and all(os.path.getsize(fi['path']) < int(args['-l']) for fi in [one, other])

                        if do_regions:
                            one_regions, other_regions = get_unique_regions(one, other)
                            if args['-n']:
                                [one_unique_bytes, other_unique_bytes] = do_map(calculate_unique_bytes, [one_regions, other_regions])
                                if args['-u'] and not any(a >= int(args['-u']) for a in [one_unique_bytes, other_unique_bytes]):
                                    continue
                            else:
                                [one_unique_bytes, other_unique_bytes] = [0,0]
                        else:
                            [one_regions, other_regions] = [[],[]]
                            [one_unique_bytes, other_unique_bytes] = [0,0]
                            

                        for file_info, regions, unique_bytes in zip([one, other], [one_regions, other_regions], [one_unique_bytes, other_unique_bytes]) :
                            print_with_uistring(file_info)

                            if args['-n'] and do_regions:
                                # count up the number of bytes determined to not be part of an "equal" region according to SequenceMatcher
                                print unique_bytes, 'unique bytes out of', os.path.getsize(file_info['path'])

                            # print every diff?
                            if regions and args['-a']:
                                if args['-m'] != None:
                                    regions = merge_regions(regions, int(args['-m']))

                                print '\n'.join(','.join([str(start), str(end)]) for (start, end) in regions)

                                #print ' '.join(','.join([str(start), str(end)]) for start, end in regions)

                                # print PE information?
                                if args['-p']:
                                    print_pe(file_info, regions)

                    # no matching file
                    else:
                        print_with_uistring(one)
                        print "No matching file found:", directory_right
                        print #newline after printing file info

                    print #newline after diff info

    if args['-c']:
        map(shutil.rmtree, directories)
