# (c) 2013-2014 The MITRE Corporation. ALL RIGH(TS|Touch(S|s)mart) RESERVED.
# Contributions by Sam Cornwell, John Butterworth, Corey Kallenberg, and Xeno Kovah

"""
Usage:
  protections.py (per-version | per-file [-s]) [--txt] [--tpm] [-id] [-r RESULTS] [--pr-relaxed] [--bioswe-lock-disabled] ([-k KNOWN]|[-g GENERATED]) [-o FILE] <csv-directory> ...
  protections.py (-h | --help)

Description:
  Determine SMRAM and BIOS writability based on the CSV files generated by Copernicus.
  Attempts to parse all CSV files in the given CSV directories.
  
Options:
  per-version        aggregate output on a product name and bios vendor/version basis
  per-file           aggregate output on a input file basis
  --pr-relaxed       do not require that the PR registers cover the entire BIOS region
  -s                 include column for the source file
  -r RESULTS         print the requested result values from parsing the CSV files.  Example: -r smm_bwp,d_lck.  Use "-r all" to print all results.
  -i                 filter out machines with inconsistent protections across multiple CSV files
  -d                 deduplicate by hostname, processing only the newest file from any particular host
  -k KNOWN           file containing list of known upgradeable-to-locked-BIOS product names
  -g GENERATED       if given, will use and output generated list of known upgradeable-to-locked-BIOS product names to this file
  -o, --output FILE  if given, will output the results into a CSV-formatted file
  --txt              include TXT (Intel Trusted Execution Technology) information
  --tpm              include TPM (Trusted Platform Module) information
"""

from docopt import docopt

import sys
import os
import glob
import ast
import string
from collections import defaultdict
import itertools
from copernicus_csv import Version, csv_to_dict, parse_all_csvs, group_by_version
import re

PR_WRITE_PROTECT = 31
COP_HDATA_SMRR_UNSUPPORTED = 12

master_upgradeable_set = set()

def init_master_upgradeable_set(filename):
    try:
        global master_upgradeable_set
        master_upgradeable_set = {line.strip() for line in open(filename, 'r').readlines()}
        return True
    except Exception, e:
        return False
    
def write_master_upgradeable_set(filename):
    try:
        with open(filename, 'w') as f:
            f.write('\n'.join(sorted(list(master_upgradeable_set))) + '\n' if master_upgradeable_set else '')
        return True
    except Exception, e:
        return False

def add_to_master_upgradeable_set(product_name):
    master_upgradeable_set.add(product_name)
            
def inferred_upgradeable_product(product, aggregated_data):
    return aggregated_data.bios_unlocked not in [0, aggregated_data.count]

def listed_upgradeable_product(product, aggregated_data):
    return product in master_upgradeable_set
 
class ParsedData(object):
    pr_relaxed = False
    bioswe_lock_disabled = False

    protections = [
        'd_lck',
        'smrr_enabled',
        'prs_cover_bios',
        'flash_lockdown',
        'bioswe_lock',
        'smm_bwp',
        'smi_lock',
    ]

    results = protections + ['vu255726', 'vu912156', 'vu552286', 'fregs_have_gaps']

    txt_and_tpm = ['feature_control_unlocked', 'senter_global_enable', 'tpm_present', 'tpm_enabled', 'tpm_activated']

    def __init__(self, values, filename = None):
        self.values = values
        self.filename = filename
        self.d_lck = d_lck_set(values)
        self.smrr_enabled = smrr_enabled(values)
        self.prs_cover_bios = prs_cover_bios(values, ParsedData.pr_relaxed)
        self.flash_lockdown = bool(int(values.get('FLOCKDN', 1)))
        self.bioswe_lock = bool_str_to_bool(values['BIOSWE_LOCK']) and not ParsedData.bioswe_lock_disabled
        self.smm_bwp = smm_bwp_set(values)
        self.smi_lock = smi_lock_set(values)
        self.feature_control_unlocked = not bit_is_set(values['IA32_FEATURE_CONTROL_MSR'], 0)
        self.senter_global_enable = bit_is_set(values['IA32_FEATURE_CONTROL_MSR'], 15)
        self.feature_control = int(values['IA32_FEATURE_CONTROL_MSR'], 16)
        self.tpm_present = bool_str_to_bool(values.get('TPM_PRESENT', 'False'))
        self.tpm_enabled = bool_str_to_bool(values.get('TPM_ENABLED', 'False'))
        self.tpm_activated = bool_str_to_bool(values.get('TPM_ACTIVATED', 'False'))
    
    # legacy dell capsule update vulns
    @property
    def vu912156(self):
        return vuln_to_cert_912156(Version(self.values))

    # gambits (UEFI reference code capsule vulns)
    @property
    def vu552286(self):
        return vuln_to_cert_552286(Version(self.values))

    # cache poisoning vuln
    @property
    def vu255726(self):
        return not self.smrr_enabled and not self.pr_protects_bios

    @property
    def smram_locked(self):
        return self.smrr_enabled

    @property
    def smram_unlocked(self):
        return not self.smram_locked

    @property
    def pr_protects_bios(self):
        return self.prs_cover_bios and self.flash_lockdown

    @property
    def uses_pr_write_protect(self):
        return any (
            bit_is_set(self.values[pr], PR_WRITE_PROTECT)
            for pr in ['PR0', 'PR1', 'PR2', 'PR3', 'PR4']
        )

    @property
    def supports_smm_bwp(self):
        return 'PCH' in [self.values.get('CHIPSET_FAMILY'), self.values.get('ICH_FAMILY')]

    @property
    def fregs_have_gaps(self):
        return bool(freg_gaps(self.values))

    '''
    @property
    def bios_locked(self):
        return
        (
            self.pr_protects_bios
            or
            (
                self.smram_locked
                and
                (
                    (self.bioswe_lock and self.smi_lock)
                    or
                    self.smm_bwp
                )
            )
        )
    ''' 
    @property
    def bios_locked(self):
        return self.pr_protects_bios or (self.smram_locked and ((self.bioswe_lock and self.smi_lock) or self.smm_bwp))

    @property
    def bios_unlocked(self):
        return not self.bios_locked


    def equal_protections(self, other):
        return all(getattr(self, p) == getattr(other, p) for p in ParsedData.protections)


class AggregatedProtections(object):
    results = ParsedData.results + ['bios_unlocked', 'smram_unlocked'] + ParsedData.txt_and_tpm

    def __init__(self, machines = None, version = None):
        self.count = 0

        for attr in AggregatedProtections.results:
            setattr(self, attr, 0)
        
        if machines:
            self.aggregate(machines)

        if version:
            self.version = version

    # all the variables casted as ints are originally bools, so they will be 1 (True) or 0 (False)
    def add_protections(self, data):
        # allows us aggregate merge sets already-aggregated machines
        if hasattr(data, 'count'):
            self.count += data.count
        else:
            self.count += 1

        for attr in AggregatedProtections.results:
            attr_sum = getattr(self, attr) + getattr(data, attr)
            setattr(self, attr, attr_sum)

    def aggregate(self, machines):
        for machine in machines:
            self.add_protections(machine)

def hex_str_to_long(s):
    #return reduce(lambda acc, (i, c): acc | (ord(c) << 8*i), enumerate(s.decode('hex')[::-1]), 0)
    return int(s, 16)

def bool_str_to_bool(s):
    return ast.literal_eval(s)
    
def get_bit_field(value, bottom, num_bits):
    if isinstance(value, (int, long)):
        int_value = value
    elif isinstance(value, str):
        int_value = hex_str_to_long(value)

    mask = ((1 << num_bits) - 1) << bottom
    return (int_value & mask) >> bottom

def bit_is_set(value, bit):
    return bool(get_bit_field(value, bit, 1))

def sanitize_region((base, limit)):
    if (base > limit):
        return None
    else:
        return base, limit

# works for both PR and FREG registers
def region_str_to_region(reg_str):
    reg_value = hex_str_to_long(reg_str)
    base = (reg_value & 0x00001fff) << 12
    limit = (((reg_value & 0x1fff0000) >> 4) | 0x00000fff)
    return sanitize_region((base, limit))

def region_intersect((base1, limit1), (base2, limit2)):
    sanitize_region((max(base1, base2), min(limit1, limit2)))

#this actually isn't necesary anymore, can use region_set_union instead
def region_merge((base1, limit1), (base2, limit2)):
    # this could probably be optimized, but it's easier to create the algorithm
    # by simply eliminating every case one at a time
    # the cases where one region is completely in the other
    if base1 >= base2 and limit1 <= limit2:
        return [(base2, limit2)]
    if base2 >= base1 and limit2 <= limit1:
        return [(base1, limit1)]
    # the cases where regions are completely separated
    if limit1 < base2 - 1 or limit2 < base1 - 1:
        return [(base1, limit1), (base2, limit2)]
    # finally the cases where the regions partially overlap
    if base1 < base2:
        return [(base1, limit2)]
    else:
        return [(base2, limit1)]

def region_cmp((base1, limit1), (base2, limit2)):
    return cmp(base1, base2) if base1 != base2 else cmp(limit1, limit2)

def region_set_union(regions):
    regions = sorted(regions, cmp=region_cmp)
    merged_regions = []
    while len(regions) >= 2:
        base1, limit1 = regions.pop(0)
        base2, limit2 = regions[0]
        
        if (base1 == base2):
            continue
        
        if (limit1 >= base2 - 1):
            regions[0] = (base1, max(limit1, limit2))
        else:
            merged_regions.append((base1, limit1))
        
    if (regions):
        merged_regions.append(regions.pop(0))
    
    return merged_regions

# region1 - region2
# warning: reading may cause aneurysm
def region_diff((base1, limit1), (base2, limit2)):
    if base1 < base2:
        if limit2 < limit1:
            return [(base1, base2 - 1), (limit2 + 1, limit1)]
        else:
            return [(base1, min(base2 - 1, limit1))]
    else:
        if limit2 >= limit1:
            return []
        else:
            return [(max(limit2 + 1, base1), limit1)]

def region_set_diff(region_set, region2):
    return region_set_union(sum([region_diff(region1, region2) for region1 in region_set], []))
    
def freg_gaps(values):
    if not bool_str_to_bool(values['FLASH_DESCRIPTOR_MODE']):
        return []

    fregs = ['FREG' + n for n in map(str, range(5))]

    flash_regions = [(0, hex_str_to_long(values['FLASH_CHIP_SIZE']) - 1)]

    for freg in fregs:
        # parse out the base, limit from the PR register
        freg_region = region_str_to_region(values[freg])

        # the region could be invalid (unused)
        if not freg_region:
            continue
        
        # subtract off the protected region from the bios region parts
        # which have not been covered yet by another PR
        flash_regions = region_set_diff(flash_regions, freg_region)

    return flash_regions

def prs_cover_bios(values, relaxed=False):
    # these all have pr0 - pr2
    if get_ich_family(values) in ['ICH7', 'ICH8', 'ICH9', 'ICH10', 'PCH']:
        pr_set = ['PR0', 'PR1', 'PR2']
        # all except ICH7 have pr3 and pr4 as well
        if get_ich_family(values) != 'ICH7':
            pr_set += ['PR3', 'PR4']
    else:
        # we don't know what PRs are available, if any, so just return False
        return False

    # I put bios_range into a list to start because it will potentially get split up into a list of pieces
    # BIOS region is irrelevant if we are not in relaxed mode
    if not relaxed:
        if bool_str_to_bool(values['FLASH_DESCRIPTOR_MODE']):
            bios_regions = [region_str_to_region(values['FREG1'])]
        else:
            bios_regions = [(0, hex_str_to_long(values['FLASH_CHIP_SIZE']) - 1)]
        
    for pr in pr_set:
        # check if this PR offers write-protection
        if not bit_is_set(values[pr], PR_WRITE_PROTECT):
            continue
        elif relaxed:
            # if the PR offers write-protection, then that's good enough from a relaxed perspective!
            return True
        
        # parse out the base, limit from the PR register
        pr_region = region_str_to_region(values[pr])
        # the region could be invalid (unused)
        if not pr_region:
            continue
        
        # subtract off the protected region from the bios region parts
        # which have not been covered yet by another PR
        bios_regions = region_set_diff(bios_regions, pr_region)

    if relaxed:
        return False
    else:
        # the bios region set should be empty if it has been completely covered by PRs
        return not bios_regions

def get_ich_family(values):
    return values.get('ICH_FAMILY', values.get('CHIPSET_FAMILY'))

def smm_bwp_set(values):
    # this bit was first seen in the Intel 5 series, which is also the first version with a PCH
    return get_ich_family(values) == 'PCH' and bit_is_set(values['BIOS_CNTL'], 5)

def smi_lock_set(values):
    try:
        return bit_is_set(values['GEN_PMCON_1'], 4)
    except:
        # return True by default in order to support older data where GEN_PMCON_1 was not collected
        return True
    
def d_lck_set(values):
    return bit_is_set(values.get('SMRAMC', 0), 4)

# HDATA_SMRR_UNSUPPORTED is deprecated, but leaving it here for now
def smrr_supported(values):
    return (bool_str_to_bool(values.get('SMRR_SUPPORTED', 'True')) and
        not bit_is_set(values['COP_RETURN_STATUS'], COP_HDATA_SMRR_UNSUPPORTED))
    
def smrr_enabled(values):
    return smrr_supported(values) and bit_is_set(values.get('SMRR_PHYSMASK', 0), 11)

def aggregate_by_product_name(parsed_data):
    groups = defaultdict(AggregatedProtections)
    
    for p in parsed_data:
        groups[p.values['PRODUCT_NAME'].strip()].add_protections(p)
    
    return groups

def group_by_hostname(parsed_data):
    groups = defaultdict(list)
    
    for p in parsed_data:
        groups[p.values.get('HOSTNAME', '')].append(p)
    
    return groups

def timestamp_from_parsed_data(parsed_data):
    try:
        tail = os.path.split(parsed_data.filename)[1]
        root = os.path.splitext(tail)[0]
        return int(root.split('=')[1], 16)
    except:
        return 0

def filter_inconsistent_machines(parsed_data):
    groups = group_by_hostname(parsed_data)
    if '' in groups:
        del groups['']
    
    for hostname, group in groups.iteritems():
        # I want to check here that the protections are all the same, so I compare them all to the first in the list
        if [p for p in group if not group[0].equal_protections(p)] != []:
            parsed_data = [p for p in parsed_data if p not in group]
    
    return parsed_data

def deduplicate_by_hostname(parsed_data):
    groups = group_by_hostname(parsed_data)
    no_hostname = groups.pop('', [])
    return [ sorted(group, key = timestamp_from_parsed_data, reverse = True)[0] for group in groups.itervalues()] + no_hostname

def get_max_len(arr):
    return max(map(len, arr))

def pad_to_max_len(s, arr):
    return s.ljust(get_max_len(arr), ' ')

VENDOR_ID_MAP = defaultdict(str, {  
    "1022" : "AMD",
    "1114" : "Atmel",
    "14e4" : "Broadcom",
    "1014" : "IBM",
    "15d1" : "Infineon",
    "8086" : "Intel",
    "17aa" : "Lenovo",
    "100b" : "National Semi",
    "1b4e" : "Nationz", 
    "1011" : "Qualcomm",
    "1055" : "SMSC", 
    "104a" : "STMicroelectrics",
    "144d" : "Samsung",
    "19fa" : "Sinosun",
    "104c" : "Texas Instruments",
    "1050" : "Nuvoton Technology"
})

def tpm_vendor_id_str(row_data, version):
    if row_data.tpm_present:
        return version.tpm_vendor + ' (' + VENDOR_ID_MAP[version.tpm_vendor] + ')'
    else:
        return ''

CERT_912156_MAP = {
    "Latitude D530" : "A12",
    "Latitude D531" : "A12",
    "Latitude D630" : "A19",
    "Latitude D631" : "A12",
    "Latitude D830" : "A17",
    "Precision M2300" : "A11",
    "Precision M4300" : "A17",
    "Precision M6300" : "A15",
    "Latitude E5400" : "A19",
    "Latitude E5500" : "A19",
    "Latitude E4200" : "A24",
    "Latitude E4300" : "A26",
    "Latitude E6400" : "A34",
    "Latitude XT2" : "A15",
    "Latitude E6500" : "A29",
    "Latitude Z600" : "A11",
    "Precision M2400" : "A28",
    "Precision M4400" : "A29",
    "Precision M6400" : "A13",
    "Precision M6500" : "A10"
}

def vuln_to_cert_912156(version):
    for product_name, bios_version in CERT_912156_MAP.iteritems():
        if product_name in version.product_name:
            try:
                if version.bios_version[0] != bios_version[0] or int(version.bios_version[1:]) < int(bios_version[1:]):
                    return True
                else:
                    return False
            except:
                return True
    return False

def hp_f_version_key(value):
    return int(re.split('\S*F\.([\dABCDEFabcdef]+)', value)[-2], 16)

def dell_version_key(value):
    return int(''.join(c for c in value if c in '0123456789'))

def re_or(re_list):
    return '|'.join( '(' + item + ')' for item in re_list)

#####
CERT_552286_MAP = {
    # Dell
    re.compile(re_or([
        'OptiPlex 3020',
    ])) : {"min_version" : "A05", "version_key" : dell_version_key},

    re.compile(re_or([
        'OptiPlex 3011 AIO',
        'PowerEdge Server T20',
    ])) : {"min_version" : "A06", "version_key" : dell_version_key},

    re.compile(re_or([
        'OptiPlex 7020',
        'OptiPlex 9020',
        'Precision WorkStation R7610'
    ])) : {"min_version" : "A08", "version_key" : dell_version_key},

    re.compile(re_or([
        'Latitude 6430U',
        'Latitude E5440',
        'Latitude E5540',
        'Latitude E6440',
        'OptiPlex 9020 AIO',
        'Precision WorkStation T3610',
        'Precision WorkStation T5610',
        'Precision WorkStation T7610',
    ])) : {"min_version" : "A09", "version_key" : dell_version_key},

    re.compile(re_or([
        'Precision WorkStation T1700',
        'Precision WorkStation M6800',
        'Precision WorkStation M4800',
    ])) : {"min_version" : "A11", "version_key" : dell_version_key},

    re.compile(re_or([
        'Latitude E6540',
        'Latitude E7240',
        'Latitude E7440',
    ])) : {"min_version" : "A12", "version_key" : dell_version_key},

    re.compile(re_or([
        'OptiPlex 3010',
        'Precision Mobile WorkStation M4700',
        'Latitude E7440',
    ])) : {"min_version" : "A13", "version_key" : dell_version_key},

    re.compile(re_or([
        'Latitude E6230',
        'Latitude E6330',
        'Latitude E6430S',
        'Precision Mobile WorkStation M6700',
    ])) : {"min_version" : "A14", "version_key" : dell_version_key},

    re.compile(re_or([
        'Latitude E5530',
        'Latitude E5430',
    ])) : {"min_version" : "A15", "version_key" : dell_version_key},

    re.compile(re_or([
        'Latitude E6530',
        'Latitude E6430',
        'OptiPlex 9010 AIO',
    ])) : {"min_version" : "A16", "version_key" : dell_version_key},

    re.compile(re_or([
        'Precision WorkStation T1650',
    ])) : {"min_version" : "A18", "version_key" : dell_version_key},

    re.compile(re_or([
        'OptiPlex 7010',
        'OptiPlex 9010',
    ])) : {"min_version" : "A19", "version_key" : dell_version_key},

    # HP
    re.compile(re_or([
        'EliteBook 8440(p|w)',
        'ProBook 6440b',
        'ProBook 6540b',
        'EliteBook 8540(p|w)',
        'ProBook 6450b',
        'ProBook 6550b',
        'ProBook 6445b',
        'ProBook 6545b',
        'EliteBook 2540p',
        'ProBook 6455b',
        'ProBook 6555b',
        'EliteBook 2740p',
        'EliteBook 8740w',
        '4320t Mobile Thin Client',
        'Mini 5102',
        'Mini 5103',
        'ProBook 4520s',
        'ProBook 4720s',
        'ProBook 4320s',
        'ProBook 4321s',
        'ProBook 4421s',
        'ProBook 4525s',
        'ProBook 4325s',
        'ProBook 4326s',
        'ProBook 4425s',
        'ProBook 5220m',
        'ProBook 5320m',
    ])) : {"min_version" : "F.50", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 500-a\w{2}',
        'Envy 700-0\w{2}',
    ])) : {"min_version" : "80.51", "version_key" : float},

    re.compile(re_or([
        'Envy Phoenix 810-0\w{2}',
        'Envy Phoenix 800-0\w{2}',
        'Envy 700-0\w{2}',
    ])) : {"min_version" : "80.21", "version_key" : float},

    re.compile(re_or([
        'CQ100-0\w{2} Desktop',
        '110-0\w{2} Desktop',
    ])) : {"min_version" : "80.16", "version_key" : float},

    re.compile(re_or([
        'Pavilion Slimline 400-0\w{2}',
        'Pavilion 500-0\w{2}',
        'Envy Phoenix 810-0\w{2}',
    ])) : {"min_version" : "80.15", "version_key" : float},

    re.compile(re_or([
        'Pavilion Slimline 400-0\w{2}',
        'CQ 100-0\w{2} Desktop',
        'Envy Recline 23-\w{3} (AiO|All-in-One)',
        'Envy Recline 23-\w{3} (TS|Touch(S|s)mart) Beats SE (AiO|All-in-One)',
        '110-0\w{2} Desktop',
        'Envy 27-\w{3} Recline (AiO|All-in-One)',
    ])) : {"min_version" : "80.13", "version_key" : float},

    re.compile(re_or([
        '23\w{0,1} Omni \w{3}',
        '23-\w{4} (AiO|All-in-One)',
        '23\w{0,1} (TS|Touch(S|s)mart) 23-f2\w{2}',
        '23\w{0,1} (TS|Touch(S|s)mart) 23-f2\w{2}',
    ])) : {"min_version" : "80.1", "version_key" : float},
     
    re.compile(re_or([
        '23-g0\w{2}',
        '21-h0\w{2}',
        '23-g0\w{2}',
        '23-g0\w{2}',
        'Pavilion 23-p0\w{2} (AiO|All-in-One)',
    ])) : {"min_version" : "80.09", "version_key" : float},

    re.compile(re_or([
        'CQ 18-40\w{2}',
        '18-50\w{2}',
        '205 G1 (AiO|All-in-One) Business',
    ])) : {"min_version" : "80.08", "version_key" : float},

    re.compile(re_or([
        '20\w{0,1} \w{3}',
        '18-\w{4} (AiO|All-in-One)',
        '1155 (AiO|All-in-One) Business',
        '20\w{0,1} (TS|Touch(S|s)mart) 20-f2\w{2}',
        '18-\w{4} (AiO|All-in-One)',
        '20\w{0,1} (TS|Touch(S|s)mart) 23-f2\w{2}',
        'Envy 23-d\w{3} (TS|Touch(S|s)mart) (AiO|All-in-One)',
    ])) : {"min_version" : "80.07", "version_key" : float},

    re.compile(re_or([
        '\w{4} (AiO|All-in-One)',
        'Envy 23-n\w{3} (AiO|All-in-One) PC Beats SE',
        '19-10\w{2}',
        '20-20\w{2}',
        '20-20\w{2}',
    ])) : {"min_version" : "80.05", "version_key" : float},

    re.compile(re_or([
        '23-h0\w{2}',
        '23-h0\w{2}',
    ])) : {"min_version" : "80.04", "version_key" : float},

    re.compile(re_or([
        '\w{3}',
        'Envy \w{4} (AiO|All-in-One)',
    ])) : {"min_version" : "80.02", "version_key" : float},

    re.compile(re_or([
        'CQ2\w{3} Desktop',
        'CQ2\w{3}',
        'p2-\w{4} Desktop',
        'Pavilion p2-\w{4}',
    ])) : {"min_version" : "8.23", "version_key" : float},

    re.compile(re_or([
        'Envy h8-\w{4} Desktop',
        'Pro 3515 Series MT',
        'Pro 33\w{2} Series SFF',
        'Pavilion p6',
        'Pavilion Slimline s5-X1\w{3}',
        'Pavilion p6-\w{4} Desktop',
        'Pavilion 500-0\w{2}',
        'Pavilion p7-\w{4} Desktop',
        'Pavilion Slimline 400-0\w{2}',
        'Pavilion HPE h9-1\w{3} Phoenix',
        'Pro 34\w{2} Series MT',
        'Pavilion HPE h9-1\w{3}',
        'Pavilion HPE h8-1\w{3}',
        'Pavilion Slimline s5-\w{4}',
        'Pavilion p7-1\w{3}',
    ])) : {"min_version" : "8.19", "version_key" : float},

    re.compile(re_or([
        'Pavilion p6-1\w{3}',
        'Pro 3500 Series MT',
        'Pavilion p6-\w{4} Desktop',
    ])) : {"min_version" : "8.18", "version_key" : float},

    re.compile(re_or([
        '21\.5\w{0,1} Omni \w{3}-\w{4}',
        'Envy 27-\w{4} (AiO|All-in-One)',
        '20-\w{4} (AiO|All-in-One)',
        'Envy 23-c\w{3} (AiO|All-in-One)',
        '(TS|Touch(S|s)mart) 420-1\w{3}',
        'Envy 23 (AiO|All-in-One) w/(TS|Touch(S|s)mart)',
        'Pavilion 20-b\w{3} (AiO|All-in-One)',
        '23\w{0,1} Omni \w{3}',
        'Omni 220-1\w{3}',
        '(TS|Touch(S|s)mart) 520-1\w{3}',
        'Omni-1\w{3}',
        '21\.5\w{0,1} (TS|Touch(S|s)mart) \w{3}-\w{4}',
        'Pro (AiO|All-in-One) 3520 Business',
        'Pavilion 23-\w{4} (AiO|All-in-One)',
        '20\w{0,1} Omni \w{3}-\w{4}',
    ])) : {"min_version" : "8.16", "version_key" : float},
 
    re.compile(re_or([
        'Pavilion p2',
        'Envy 20-d\w{3} (TS|Touch(S|s)mart) (AiO|All-in-One)',
        '1105 (AiO|All-in-One) Business',
        'CQ2\w{3} Desktop',
        '600B Microtower',
    ])) : {"min_version" : "8.15", "version_key" : float},
     
    re.compile(re_or([
        'Pavilion 14-n0\d{2}',
        'Pavilion 14-n1\d{2}',
        'Pavilion 14-n0\d{2} Ultrabook',
        'Pavilion 14-n1\d{2} Ultrabook',
        'Pavilion 14-n2\d{2}',
        'Pavilion 14t-n100',
        'Pavilion 14t-n200',
        'Pavilion 15-n0\d{2}',
        'Pavilion 15-n1\d{2}',
        'Pavilion 15-n2\d{2}',
        'Pavilion 15t-n100',
        'Pavilion 15t-n200',
        'Pavilion (TS|Touch(S|s)mart) 14-n0\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n1\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n2\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n2\d{2} Ultrabook',
        'Pavilion (TS|Touch(S|s)mart) 14t-n100 Ultrabook',
        'Pavilion (TS|Touch(S|s)mart) 14t-n200 Ultrabook',
        'Pavilion (TS|Touch(S|s)mart) 15-n0\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15-n1\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15-n2\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15t-n100',
        'Pavilion (TS|Touch(S|s)mart) 15t-n200',
        'Pavilion (TS|Touch(S|s)mart) 14t-n100',
        'Pavilion (TS|Touch(S|s)mart) 14t-n200',
    ])) : {"min_version" : "F.65", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'CQ 58-bf\d{1}',
        '2000-bf\d{2}',
    ])) : {"min_version" : "F.46", "version_key" : hp_f_version_key},
          
    re.compile(re_or([
        'CQ 58-2\d{2}',
        'CQ 58-3\d{2}',
        'CQ 58t-200',
        'CQ 58t-300',
        '2000-22\d{2}',
        '2000-23\d{2}',
        '2000t-2200',
        '2000t-2300',
        '650',
        'Envy dv4-52\d{2}',
        'Envy dv4-53\d{2}',
        'Envy dv4t-5200',
        'Envy dv4t-5300',
    ])) : {"min_version" : "F.39", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 15-e0\d{2}',
        'Pavilion 15-e1\d{2}',
        'Pavilion 15z-e000',
        'Pavilion 15z-e100',
        'Pavilion 17-e0\d{2}',
        'Pavilion 17-e1\d{2}',
        'Pavilion 17z-e000',
        'Pavilion 17z-e100',
    ])) : {"min_version" : "F.33", "version_key" : hp_f_version_key},
           
    re.compile(re_or([
        'CQ 58 3\d{2}',
        'CQ 58-2\d{2}',
        'CQ 58z-200',
        'CQ 58z-300',
        '2000-22\d{2}',
        '2000-23\d{2}',
        '2000z-2200',
        '2000z-2300',
        '655',
        'Pavilion dm1-43\d{2}',
        'Pavilion dm1-44\d{2}',
        'Pavilion dm1z-4300',
        'Pavilion dm1z-4400',
        'Pavilion g4-22\d{2}',
        'Pavilion g4-23\d{2}',
        'Pavilion g4z-2200',
        'Pavilion g4z-2300',
        'Pavilion g6-22\d{2}',
        'Pavilion g6-23\d{2}',
        'Pavilion g6z-2200',
        'Pavilion g6z-2300',
        'Pavilion g7-22\d{2}',
        'Pavilion g7-23\d{2}',
        'Pavilion g7z-2200',
        'Pavilion g7z-2300',
    ])) : {"min_version" : "F.29", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'SpectreXT 13-21\d{2}',
        'SpectreXT 13-22\d{2}',
        'SpectreXT 13-23\d{2}',
        'SpectreXT 13t-2100',
        'SpectreXT 13t-2200',
        'SpectreXT 13t-2300',
    ])) : {"min_version" : "F.27", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Envy m6-11\d{2}',
        'Envy m6-12\d{2}',
        'Envy m6t-1100',
        'Envy m6t-1200',
        'Pavilion g4-22\d{2}',
        'Pavilion g4-23\d{2}',
        'Pavilion g4t-2200',
        'Pavilion g4t-2300',
        'Pavilion g6-22\d{2}',
        'Pavilion g6-23\d{2}',
        'Pavilion g6t-2200',
        'Pavilion g6t-2300',
        'Pavilion g7-22\d{2}',
        'Pavilion g7-23\d{2}',
        'Pavilion g7t-2200',
        'Pavilion g7t-2300',
        'Split 13-m0\d{2} x2',
        'Split 13-m1\d{2} x2',
        'Split 13-m2\d{2} x2',
        'Split 13t-m000 x2',
        'Split 13t-m100 x2',
        'Split 13t-m200 x2',
    ])) : {"min_version" : "F.26", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Envy 6t-1200 Sleekbook',
        'Envy 6t-1200 UltraBook',
        'Envy Sleekbook 6-10\d{2}',
        'Envy Sleekbook 6-12\d{2}',
        'Envy Sleekbook 6t-1100',
        'Envy (TS|Touch(S|s)mart) 4-12\d{2} Sleekbook',
        'Envy (TS|Touch(S|s)mart) 4t-1200 Sleekbook',
        'Envy (TS|Touch(S|s)mart) 4t-1200 UltraBook',
        'Envy UltraBook 6t-1100',
        'Envy 4t-1200 Sleekbook',
        'Envy 4t-1200 UltraBook',
        'Envy Sleekbook 4-11\d{2}',
        'Envy Sleekbook 4-12\d{2}',
        'Envy Sleekbook 4t-1100',
    ])) : {"min_version" : "F.25", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 14-a0\d{2}',
        'Pavilion 14-e1\d{2}',
        'Pavilion 14t-a000',
        'Pavilion 14t-e100',
        'Pavilion 15-e0\d{2}',
        'Pavilion 15-e1\d{2}',
        'Pavilion 15t-e000',
        'Pavilion 15t-e100',
        'Pavilion 17-e0\d{2}',
        'Pavilion 17-e1\d{2}',
        'Pavilion 17t-e000',
        'Pavilion 17t-e100',
    ])) : {"min_version" : "F.23", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 11-h0\d{2} x2',
        'Pavilion 11-h1\d{2} x2',
        'Pavilion 11t-1000 x2',
        'Pavilion 11t-h000 x2',
        'Pro x 2 410 G1',
    ])) : {"min_version" : "F.22", "version_key" : hp_f_version_key},
           
    re.compile(re_or([
        'Pavilion (TS|Touch(S|s)mart) 15z-b000 Sleekbook',
        'Pavilion 15-b1\d{2} Sleekbook',
        'Pavilion Sleekbook 14z-1000',
        'Pavilion Sleekbook 14z-b100',
        'Pavilion (TS|Touch(S|s)mart) 15-b1\d{2} Sleekbook',
        'Pavilion Sleekbook 14-10\d{2}',
        'Pavilion Sleekbook 14-b1\d{2}',
    ])) : {"min_version" : "F.1D", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Spectre 13-h2\d{2} x2',
        'Spectre 13t-h200',
        'Split 13-f0\d{2} x2',
        'Split 13-g1\d{2} x2',
        'Split 13-g2\d{2} x2',
        'Split 13t-f000 x2',
        'Split 13t-g100 x2',
        'Split 13t-g200 x2',
        'Pavilion 14-b100 - 14-b199 Sleekbook',
        'Pavilion 15-b1\d{2} Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 14-b1\d{2} Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 15-b1\d{2} Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 14t-b100 Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 14t-b100 Ultrabook',
        'Pavilion 14t-b100 Sleekbook',
        'Pavilion 14t-b100 Ultrabook',
        'Pavilion 15t-b100 Sleekbook',
        'Pavilion 15t-b100 Ultrabook',
        'Pavilion Sleekbook 14-b0\d{2}',
        'Pavilion Sleekbook 14t-b000',
        'Pavilion Sleekbook 15-b0\d{2}',
        'Pavilion Sleekbook 15t-b000',
        'Pavilion (TS|Touch(S|s)mart) 15t-b100 Sleekbook',
        'Pavilion Ultrabook 14-b0\d{2}',
        'Pavilion Ultrabook 14t-b000',
        'Pavilion Ultrabook 15-b0\d{2}',
        'Pavilion Ultrabook 15t-b000',
    ])) : {"min_version" : "F.1C", "version_key" : hp_f_version_key},

    re.compile(re_or([
        '14-s0\d{2}',
        '14-s0\d{2} (TS|Touch(S|s)mart)',
        '14t-s000',
        '14t-s000 (TS|Touch(S|s)mart)',
        '15-s0\d{2}',
        '15-s0\d{2} (TS|Touch(S|s)mart)',
        '15t-s000',
        '15t-s000 (TS|Touch(S|s)mart)',
        '14-r0\d{2}',
        '14-r0\d{2} (TS|Touch(S|s)mart)',
        '14t-r000',
        '14t-r000 (TS|Touch(S|s)mart)',
        '15-r0\d{2}',
        '15-r0\d{2} (TS|Touch(S|s)mart)',
        '15t-r000',
        '15t-r000 (TS|Touch(S|s)mart)',
        '240 G3',
        '250 G3',
        'Envy 14-k0\d{2} Sleekbook',
        'Envy 14-k1\d{2} Sleekbook',
        'Envy 14t-k000 Sleekbook',
        'Envy 14t-k100 Sleekbook',
        'Envy (TS|Touch(S|s)mart) 14-k000 Sleekbook',
        'Envy (TS|Touch(S|s)mart) 14-k000 Ultrabook',
        'Envy (TS|Touch(S|s)mart) 14-k100 Sleekbook',
        'Envy (TS|Touch(S|s)mart) 14-k1\d{2} Ultrabook',
        'Envy (TS|Touch(S|s)mart) 14t-k100 Ultrabook',
        'Envy (TS|Touch(S|s)mart) m6-k000 Ultrabook',
        'Envy (TS|Touch(S|s)mart) m6-k100 Ultrabook',
        'Envy (TS|Touch(S|s)mart) m6t-k000 SleekBook',
        'Envy (TS|Touch(S|s)mart) m6t-k100 SleekBook',
        'Envy (TS|Touch(S|s)mart) 14-k0\d{2} Sleekbook',
        'Envy (TS|Touch(S|s)mart) 14-k0\d{2} Ultrabook',
        'Envy (TS|Touch(S|s)mart) 14-k1\d{2} SleekBook',
        'Envy (TS|Touch(S|s)mart) m6-k0\d{2} SleekBook',
        'Envy (TS|Touch(S|s)mart) m6-k0\d{2} Ultrabook',
        'Envy (TS|Touch(S|s)mart) m6-k1\d{2} SleekBook',
        'Envy (TS|Touch(S|s)mart) m6-k1\d{2} Ultrabook',
        'Envy m4-10\d{2}',
        'Envy m4-11\d{2}',
        'Envy m4t-1000',
        'Envy m4t-1100',
        'Envy m6-k000 SleekBook',
        'Envy m6-k0\d{2} SleekBook',
    ])) : {"min_version" : "F.18", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        '14-a0\d{2}',
        '14-a0\d{2} (TS|Touch(S|s)mart)',
        '14t-a000',
        '14t-a000 (TS|Touch(S|s)mart)',
        '15-a0\d{2}',
        '15-a0\d{2} (TS|Touch(S|s)mart)',
        '15t-a000',
        '15t-a000 (TS|Touch(S|s)mart)',
        'Envy Sleekbook 6-10\d{2}',
        'Envy Sleekbook 6-11\d{2}',
        'Envy Sleekbook 6-12\d{2}',
        'Envy Sleekbook 6z-1000',
        'Envy Sleekbook 6z-1200',
        '14-d0\d{2}',
        '14-d0\d{2} (TS|Touch(S|s)mart)',
        '14t-d000',
        '14t-d000 (TS|Touch(S|s)mart)',
        '15-d0\d{2}',
        '15-d0\d{2} (TS|Touch(S|s)mart)',
        '15t-d000',
        '15t-d000 (TS|Touch(S|s)mart)',
        '240 G2',
        '246 G2',
        '250 G2',
        'Spectre 14-32\d{2}',
        'Spectre 14-33\d{2}',
        'Spectre 14t-3200',
        'Spectre 14t-3300',
    ])) : {"min_version" : "F.17", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        '14-h0\d{2}',
        '14-h0\d{2} (TS|Touch(S|s)mart)',
        '14z-h000',
        '14z-h000 (TS|Touch(S|s)mart)',
        '15-h0\d{2}',
        '15-h0\d{2} (TS|Touch(S|s)mart)',
        '15z-h000',
        '15z-h000 (TS|Touch(S|s)mart)',
        '14-g0\d{2}',
        '14-g0\d{2} (TS|Touch(S|s)mart)',
        '14z-g000',
        '15-g0\d{2}',
        '15-g0\d{2} (TS|Touch(S|s)mart)',
        '15z-g000',
        '15z-g000 (TS|Touch(S|s)mart)',
        '245 G3',
        '255 G3',
        'Envy 15-j0\d{2}',
        'Envy 15-j1\d{2}',
        'Envy 15z-j000',
        'Envy (TS|Touch(S|s)mart) 15-j0\d{2}',
        'Envy (TS|Touch(S|s)mart) 15z-j000',
        'Envy (TS|Touch(S|s)mart) 15z-j100',
    ])) : {"min_version" : "F.16", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 11-e0\d{2}',
        'Pavilion 11-e1\d{2}',
        'Pavilion 11z-e000',
        'Pavilion 11z-e100',
        'Pavilion 14-n0\d{2}',
        'Pavilion 14-n1\d{2}',
        'Pavilion 14-n2\d{2}',
        'Pavilion 14z-n100',
        'Pavilion 14z-n200',
        'Pavilion (TS|Touch(S|s)mart) 11-e0\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 11-e1\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n0\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n1\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14-n2\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 14z-n100',
        'Pavilion (TS|Touch(S|s)mart) 14z-n200',
        'Pavilion (TS|Touch(S|s)mart) 15-n0\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15-n1\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15-n2\d{2}',
        'Pavilion (TS|Touch(S|s)mart) 15z-n100',
        'Pavilion (TS|Touch(S|s)mart) 15z-n200',
        'Pavilion (TS|Touch(S|s)mart) 11z-e000',
        'Pavilion (TS|Touch(S|s)mart) 11z-e100',
        'Pavilion 15-n0\d{2}',
        'Pavilion 15-n1\d{2}',
        'Pavilion 15-n2\d{2}',
        'Pavilion 15z-n100',
        'Pavilion 15z-n200',
        '215 G1 Ultrabook',
    ])) : {"min_version" : "F.15", "version_key" : hp_f_version_key},
      
    re.compile(re_or([
        'Envy Rove 20-k0\d{2}',
        'Envy Rove 20-k2\w{2}',
        'Spectre 13-30\d{2} Ultrabook',
        'Spectre 13t-3000 Ultrabook',
    ])) : {"min_version" : "F.14", "version_key" : hp_f_version_key},
 
    re.compile(re_or([
        '248 G1',
        '340 G1',
        '350 G1',
        'Pavilion 11-n0\d{2} x360',
        'Pavilion 11t x360',
        'SpectreXT (TS|Touch(S|s)mart) 15-41\d{2}',
        'SpectreXT (TS|Touch(S|s)mart) 15t-4000',
        'SpectreXT (TS|Touch(S|s)mart) 15t-4100',
    ])) : {"min_version" : "F.12", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Beats 15-p0\d{2}',
        'Beats 15z',
        'Pavilion 14-v0\d{2}',
        'Pavilion 14z',
        'Pavilion 15-p0\d{2}',
        'Pavilion 15z',
        'Pavilion 17-f0\d{2}',
        'Pavilion 17z',
    ])) : {"min_version" : "F.10", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        '14-a0\d{2}',
        '14-a0\d{2} (TS|Touch(S|s)mart)',
        '14z-a000',
        '14z-a000 (TS|Touch(S|s)mart)',
        '15-a0\d{2}',
        '15-a0\d{2} (TS|Touch(S|s)mart)',
        '15z-a000',
        '15z-a000 (TS|Touch(S|s)mart)',
        '14-d0\d{2}',
        '14-d0\d{2} (TS|Touch(S|s)mart)',
        '14z-d000',
        '14z-d000 (TS|Touch(S|s)mart)',
        '15-d0\d{2}',
        '15-d0\d{2} (TS|Touch(S|s)mart)',
        '15z-d000',
        '15z-d000 (TS|Touch(S|s)mart)',
        '245 G2',
        '255 G2',
    ])) : {"min_version" : "F.0E", "version_key" : hp_f_version_key},

    re.compile(re_or([
        '242 G1',
    ])) : {"min_version" : "F.0B", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'Pavilion 14-f0\d{2} Sleekbook',
        'Pavilion 14z-f000 Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 14-f0\d{2} Sleekbook',
        'Pavilion (TS|Touch(S|s)mart) 14z-f000 Sleekbook',
    ])) : {"min_version" : "F.08", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        'Pavilion 10-e0\d{2} (TS|Touch(S|s)mart)',
        'Pavilion 10z-e000 (TS|Touch(S|s)mart)',
        'Pavilion 13-p1\d{2} x2',
        'Pavilion 13t-p100 x2',
    ])) : {"min_version" : "F.07", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        '14-w0\w{2}',
        '14t-w0\w{2}',
        '15-f0\w{2}',
        '15z-f0\w{2}',
        '242 G2',
        'Envy 15-u0\d{2} x360',
        'Envy 15t-u0\w{2} x360',
        'Pavilion 13-a0\d{2} x360',
        'Pavilion 13-b0\d{2} x360',
        'Pavilion 13-r0\w{2} x2',
        'Pavilion 13t-a0\w{2} x360',
        'Pavilion 13t-b0\w{2}',
        'Pavilion m4-10\d{2}',
        'Split 13-r0\d{2} x2',
        'Split 13-r0\w{2} x2',
        'P Pavilion 13-r0\d{2} x2',
    ])) : {"min_version" : "F.06", "version_key" : hp_f_version_key},
    
    re.compile(re_or([
        'Pavilion 13-a0\w{2} x360',
        'Pavilion 13z x360',
    ])) : {"min_version" : "F.05", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'CQ45-c\d{2}',
        '1000-C1\d{2}',
        '246',
        '345 G2',
        '355 G2',
        '450',
    ])) : {"min_version" : "F.04", "version_key" : hp_f_version_key},

    re.compile(re_or([
        'g14-a0\d{2}',
    ])) : {"min_version" : "F.02", "version_key" : hp_f_version_key},
}
#####

def vuln_to_cert_552286(version):
    for product_regex, compare_info in CERT_552286_MAP.iteritems():
        if product_regex.search(version.product_name):
            try:
                version_key = compare_info['version_key']
                min_version = compare_info['min_version']
                if version_key(version.bios_version) < min_version:
                    return True
                else:
                    return False
            except:
                pass
    return False

class Table(object):
    def __init__(self, columns=None):
        self.columns = columns
        if self.columns:
            self.num_columns = len(columns)
        else:
            self.num_columns = 0
        self.rows = []

    def add_row(self, row):
        if not self.rows and not self.num_columns:
            self.num_columns = len(row)
        elif len(row) != self.num_columns:
            raise Exception("Row does not contain matching number of columns.")
        self.rows.append(row)

    def sort_rows(self, keys):
        indices = [self.columns.index(key) for key in keys]
        sort = lambda row: tuple(row[i].lower() for i in indices)
        self.rows = sorted(self.rows, key = sort)

    def print_table(self):
        column_widths = [0] * self.num_columns
        for i in range(self.num_columns):
            rows = itertools.chain([self.columns] if self.columns else [], self.rows)
            column_widths[i] = max(len(str(row[i])) for row in rows)

        rows = itertools.chain([self.columns] if self.columns else [], self.rows)
        for row in rows:
            print ' '.join(cell_val.ljust(width, ' ') for cell_val, width in itertools.izip(row, column_widths))

    def write_table(self, out):
        if self.columns:
            out.write(','.join(self.columns) + '\n')
        for row in self.rows:
            out.write(','.join(row) + '\n')

def aggregate_protections(parsed_data):
    return [AggregatedProtections(machines, version) for version, machines in group_by_version(parsed_data).iteritems()]

def print_summary_row(count, total, message):
    if total:
        percent = str(float(count)/float(total) * 100)
    else:
        percent = 'NaN'

    print count, 'out of', total, '(' + percent + '%)', message


# MAIN
if __name__ == '__main__':
    arguments = docopt(__doc__)
    
    if arguments['--pr-relaxed']:
        ParsedData.pr_relaxed = True
    if arguments['--bioswe-lock-disabled']:
        ParsedData.bioswe_lock_disabled = True


    parsed_data = parse_all_csvs(arguments['<csv-directory>'], data_transform=ParsedData).values()

    # not sure if this should be before or after the summary calculations shortly below
    # I put it here on the off chance we remove all of the data, so that the next check will catch it
    if arguments['-i']:
        parsed_data = filter_inconsistent_machines(parsed_data)

    if not parsed_data:
        print "No valid CSV files found."
        sys.exit(0)

    if arguments['-d']:
        parsed_data = deduplicate_by_hostname(parsed_data)
    
    num_processed = len(parsed_data)
    num_smram_unlocked = sum(1 for p in parsed_data if p.smram_unlocked)
    num_bios_unlocked = sum(1 for p in parsed_data if p.bios_unlocked)
    num_vuln_to_cert_912156 = sum(1 for p in parsed_data if p.vu912156)
    num_vuln_to_cert_552286 = sum(1 for p in parsed_data if p.vu552286)
    num_vuln_to_cert_255726 = sum(1 for p in parsed_data if p.vu255726)
    num_smi_lock_not_set = sum(1 for p in parsed_data if not p.smi_lock)
    num_dont_use_pr_write_protect = sum(1 for p in parsed_data if not p.uses_pr_write_protect)
    num_dont_use_smm_bwp = sum(1 for p in parsed_data if not p.smm_bwp and p.supports_smm_bwp)
    num_support_smm_bwp = sum(1 for p in parsed_data if p.supports_smm_bwp)
    num_senter_global_enable = sum(1 for p in parsed_data if p.senter_global_enable)
    num_feature_control_unlocked = sum(1 for p in parsed_data if p.feature_control_unlocked)
  
    if arguments['--tpm']:
        num_tpm_present = sum(1 for p in parsed_data if p.tpm_present)
        num_tpm_enabled = sum(1 for p in parsed_data if p.tpm_enabled)
        num_tpm_activated = sum(1 for p in parsed_data if p.tpm_activated)

    if arguments['-r']:
        if arguments['-r'] == 'all':
            results_columns = ParsedData.results
        else:
            results_columns = [res for res in map(string.lower, arguments['-r'].split(',')) if res in ParsedData.results]
    else:
        results_columns = []
    
    # set up base column names
    col_names = [
        "BIOS_VENDOR",
        "PRODUCT_NAME",
        "BIOS_VERSION",
        "SMRAM_UNLOCKED",
        "BIOS_UNLOCKED"]
    
    # add column names specific to each command type
    if arguments['per-version']:
        col_names.insert(0, "COUNT")
    elif arguments['per-file']:
        col_names.insert(0, "HOSTNAME")
    
    # these can be added for either command
    if arguments['-r']:
        col_names.extend(res.upper() for res in results_columns)
         
    if arguments['--txt']:
        if arguments['per-version']:
            col_names.extend(["FEATURE_CONTROL_UNLOCKED", "SENTER_GLOBAL_ENABLE"])
        elif arguments['per-file']:
            col_names.extend(["FEATURE_CONTROL_UNLOCKED", "SENTER_GLOBAL_ENABLE", "FEATURE_CONTROL"])
    
    if arguments['--tpm']:
        col_names.extend(["TPM_PRESENT", "TPM_ENABLED", "TPM_ACTIVATED", "TPM_VENDOR"])
        
    # this is command-specific, but I want it to be the last column, so I add it separately here
    if arguments['per-file'] and arguments['-s']:
        col_names.append("SOURCE_FILE")
    
    try:
        outfile = open(arguments['--output'], 'w')
        outfile.write(','.join(map(string.strip, col_names)))
        outfile.write('\n')
    except IOError:
        print "Could not create output file", arguments['--output']
        outfile = None
    except Exception, e:
        # there was no output file passed
        outfile = None
    
    table = Table(col_names)

    if arguments['per-version']:
        rows = aggregate_protections(parsed_data)

    elif arguments['per-file']:
        rows = parsed_data

    vendors = {}

    for row_data in rows:
        row = []
       # add in the basic fields
        if arguments['per-version']:
            row.append(str(row_data.count))
            version = row_data.version
                        
            if arguments['--tpm'] and row_data.tpm_present:          
                if vendors.has_key(VENDOR_ID_MAP[version.tpm_vendor]):        
                    vendors[VENDOR_ID_MAP[version.tpm_vendor]] += row_data.count
                else:
                    vendors[VENDOR_ID_MAP[version.tpm_vendor]] = row_data.count 
                    
        elif arguments['per-file']:
            row.append(row_data.values.get('HOSTNAME', ''))
            version = Version(row_data.values)
                    
            if arguments['--tpm'] and row_data.tpm_present:
                if vendors.has_key(VENDOR_ID_MAP[version.tpm_vendor]):        
                    vendors[VENDOR_ID_MAP[version.tpm_vendor]] += 1
                else:
                    vendors[VENDOR_ID_MAP[version.tpm_vendor]] = 1
        
        row.extend([version.bios_vendor,
                version.product_name,
                version.bios_version,
                str(row_data.smram_unlocked),
                str(row_data.bios_unlocked)])

        # add in the individual protections if the command line argument was given
        if arguments['-r']:
            row.extend(str(getattr(row_data, res)) for res in results_columns)

        # add in the TXT relevant info if the command line argument was given
        if arguments['--txt']:
            if arguments['per-version']:
                row.extend([str(row_data.feature_control_unlocked), str(row_data.senter_global_enable)])
            elif arguments['per-file']:
                row.extend([str(row_data.feature_control_unlocked), str(row_data.senter_global_enable), str(row_data.feature_control)])
       
        # add in TPM data if the command line argument was given
        if arguments['--tpm']: 
            row.extend([str(row_data.tpm_present), str(row_data.tpm_enabled), str(row_data.tpm_activated), tpm_vendor_id_str(row_data, version)])

        # add in the source file if the command line argument was given
        if arguments['per-file'] and arguments['-s']:
            row.append(row_data.filename)

        table.add_row(row)

    sort_keys = ["BIOS_VENDOR", "PRODUCT_NAME", "BIOS_VERSION"]
    if(arguments['per-file']):
        sort_keys.append("HOSTNAME")

    if arguments['--tpm']:
       sort_keys.append("TPM_VENDOR")
       
    table.sort_rows(sort_keys)

    num_upgradeable = 0
    
    if arguments['-k']:
        init_master_upgradeable_set(arguments['-k'])

    try:
        table.write_table(open(arguments['--output'], 'w'))
    except IOError:
        print "Could not create output file", arguments['--output']
    except Exception, e:
        # there was no output file passed
        pass
    
    table.print_table()

    aggregated_data = aggregate_by_product_name(parsed_data)
    
    for product, agg in aggregated_data.iteritems():
        if arguments['-k']:
            if listed_upgradeable_product(product, agg):
                num_upgradeable += agg.bios_unlocked
        if arguments['-g']:
            if inferred_upgradeable_product(product, agg):
                add_to_master_upgradeable_set(product)
                num_upgradeable += agg.bios_unlocked
    
    if arguments['-g']:
        write_master_upgradeable_set(arguments['-g'])
    
    print #newline
    print num_processed, "CSV files successfully processed"
    print_summary_row(num_smram_unlocked, num_processed, "unlocked SMRAM")
    print_summary_row(num_bios_unlocked, num_processed, "unlocked BIOS")
    print_summary_row(num_smi_lock_not_set, num_processed, "SMI_LOCK not set")
    print_summary_row(num_dont_use_pr_write_protect, num_processed, "PR write protection NOT used")
    print_summary_row(num_dont_use_smm_bwp, num_support_smm_bwp, "SMM_BWP NOT used when supported")
    print_summary_row(num_vuln_to_cert_255726, num_processed, "vulnerable to CERT VU#255726")
    print_summary_row(num_vuln_to_cert_912156, num_processed, "vulnerable to CERT VU#912156")
    print_summary_row(num_vuln_to_cert_552286, num_processed, "vulnerable to CERT VU#552286")
    
    if arguments['--txt']:
        print_summary_row(num_feature_control_unlocked, num_processed, "Lock bit not set in IA32_FEATURE_CONTROL_MSR")
        print_summary_row(num_senter_global_enable, num_processed, "SENTER Global Enable set in IA32_FEATURE_CONTROL_MSR")
    
    if arguments['--tpm']: 
        print_summary_row(num_tpm_present, num_processed, "TPM present")
        print_summary_row(num_tpm_enabled, num_processed, "TPM enabled")
        print_summary_row(num_tpm_activated, num_processed, "TPM activated")
        print #newline
        print "TPM Vendor Distribution"
        print "-----------------------"
        for key in vendors: print key + " - " + str(vendors[key]), '(' + str(float(vendors[key])/float(num_tpm_present) * 100) + '%)'
 
    if arguments['-g'] or arguments['-k']:
        print_summary_row(num_upgradeable, num_bios_unlocked, "upgradeable from unlocked to locked BIOS")
